#include <ArduinoJson.h>
#include <string>
#include <unordered_map>
#include <vector>
#include <RHReliableDatagram.h>
#include <RH_RF95.h>
#include <SPI.h>
#include <ESP8266WiFi.h>
#include <ESP8266HTTPClient.h>

#define LORA_DELAY 4000

#define MAX_JSON_FIELD_LEN 32
#define MAX_REGISTRY_SIZE 255
#define REGISTRATION_BUF_SIZE 32
#define NUMBUCKETS 131      // Константа для инициализации хеш-таблицы.
#define COMMAND 0           // Индекс, по которому хранится команда.
#define BYTE_COUNT 1        // Индекс, по которому хранится число байт.
#define START_PAYLOAD 2     // Начало содержимого.

#define NOP 0               // Пустая команда.
#define REGISTRATION 1      // Команда регистрации.
#define CLEAR_REGISTRY 2    // Команда очистки реестра.
#define TELEMETRY 3

#define INTEGER 0x00
#define CUSTOM_FLOAT 0x40 
#define STRING 0x80
#define SPECIAL 0xC0
#define _8BIT 0x00
#define _16BIT 0x10
#define _32BIT 0x20
#define IP_ADDR 0x01
#define MAC_ADDR 0x02
#define DATE_YYYYMMDD 0x03
#define TIME_HHMMSS 0x04
#define JSON_OBJECT 0x05
#define GENERAL_TYPE_MASK 0xC0
#define SPECIAL_TYPE_MASK 0x3F
#define NUMBER_SIZE_MASK 0x30
#define DECIMAL_POINT_MASK 0x0F

#define JSON_LEVEL_UP 0xFF
#define LORA_ROUTER 1       // Адрес маршрутизатора.
#define LORA_GATEWAY 2      // Адрес ретранслятора.

RH_RF95 driver(D10, D4); // Соединить проводом пины D2 и D4.

RHReliableDatagram manager(driver, LORA_GATEWAY);

const char* ssid     = "***";
const char* password = "***";


uint8_t send_buf[RH_RF95_MAX_MESSAGE_LEN];   // Буфер на отправку LoRa.
uint8_t recv_buf[RH_RF95_MAX_MESSAGE_LEN];   // Буфер на получение.

char registry[MAX_REGISTRY_SIZE][MAX_JSON_FIELD_LEN];   // Реестр в виде массива строк. 
uint8_t registry_counter;      // Счётчик реестра.
uint8_t registry_counter_old;  // Старое значение счётчика.
std::unordered_map<std::string, uint8_t> registry_hash(NUMBUCKETS);   // Реестр в виде хеш-таблицы.
std::vector<std::string> last_registered(15);
char unpack_buf_cstr[MAX_JSON_FIELD_LEN];

// Функции для отладки реестра.
void print_registry_hash()
{
	Serial.println("Registry: ");
	for(auto iter = registry_hash.begin(); iter != registry_hash.end(); ++iter)
	{
		Serial.print(iter->first.c_str());
		Serial.print(F(": "));
		Serial.println((int)iter->second);
	}
}

void print_registry()
{
	for (int i = 0; i < registry_counter; i++)
	{
		Serial.print(i);
		Serial.print(F(": "));
		Serial.println(registry[i]);
	}
}

void process_packet()
{
	switch (recv_buf[COMMAND])
	{
		case REGISTRATION:
			register_keys();
			break;
		case CLEAR_REGISTRY:
			registry_counter = registry_counter_old = 0;
			registry_hash.clear();
			Serial.println(F("Registry cleared."));
			break;
		case TELEMETRY:
			unpack_telemetry();
			break;
	}
}

void cancel_registration()
{
	registry_counter = registry_counter_old;
	for (auto &key: last_registered)
	{
		registry_hash.erase(key);
	}
}

void register_keys()
{
	send_buf[COMMAND] = REGISTRATION;
	for (int i = START_PAYLOAD; i < recv_buf[BYTE_COUNT];)
	{
		strcpy(unpack_buf_cstr, (char*)&recv_buf[i]);
		i += strlen((char*)&recv_buf[i]) + 1;
		std::string unpack_buf_str = unpack_buf_cstr;
		if (registry_hash.count(unpack_buf_str) == 0)
		{
			if (registry_counter == MAX_REGISTRY_SIZE)
			{
				Serial.println(F("Registry full."));
				force_registry_clear();
				return;
			}
			strcpy(registry[registry_counter], unpack_buf_cstr);		// Если такого ключа ещё нет, то
			registry_hash.insert({unpack_buf_str, registry_counter});   // он заносится в хеш-таблицу,
			send_buf[send_buf[BYTE_COUNT]++] = registry_counter++;  // а счётчик реестра увеличивается на 1.
			last_registered.push_back(unpack_buf_str);
		}
		else
		{
			send_buf[send_buf[BYTE_COUNT]++] = registry_hash[unpack_buf_str]; // В противном случае отправляется номер
		}                                                                 // уже существующего ключа.
	}
}

void force_registry_clear()
{
    send_buf[COMMAND] = CLEAR_REGISTRY;
    send_buf[BYTE_COUNT] = START_PAYLOAD;

    registry_counter = registry_counter_old = 0;
    registry_hash.clear();
    Serial.println(F("\nRequesting registry clear."));

    for(;;)
    {
        if (manager.available())
        {
            Serial.println(F("."));

            uint8_t len = sizeof(recv_buf);
            uint8_t from;

            if (manager.recvfromAck(recv_buf, &len, &from) && from == LORA_ROUTER)
            {
                if (manager.sendtoWait(send_buf, send_buf[BYTE_COUNT], from))
                {
                    break;
                }
            }
        }
        delay(1);
    }

    Serial.println(F("\nRegistry cleared."));
}

void setup()
{
	Serial.begin(9600);
	while (!Serial);
    WiFi.mode(WIFI_STA);
    WiFi.begin(ssid, password);

    Serial.print(F("\nConnecting "));
    while (WiFi.status() != WL_CONNECTED)
    {
        delay(500);
        Serial.print(F("."));
    }
    Serial.println();

    Serial.print(F("Connected. IP address: "));
    Serial.println(WiFi.localIP());
    Serial.print(F("SSID: "));
    Serial.println(WiFi.SSID());

	while (!manager.init())
	{
		Serial.println(F("Init failed."));
	}
	manager.setTimeout(LORA_DELAY);
	driver.setFrequency(868.9);
	driver.setTxPower(13, false);

	force_registry_clear();
}

void loop()
{
	send_buf[COMMAND] = NOP;
	send_buf[BYTE_COUNT] = START_PAYLOAD;

	registry_counter_old = registry_counter;

	if (manager.available())
	{
		uint8_t len = sizeof(recv_buf);
		uint8_t from;

		if (manager.recvfromAck(recv_buf, &len, &from) && from == LORA_ROUTER)
		{
			Serial.println(F("Got request from LoRa router."));
			process_packet();

			if (!manager.sendtoWait(send_buf, sizeof(send_buf), from))	// Если пакет не дошёл,
			{															// и если отправляемая команда - запрос на регистрацию, то отменяем регистрацию.
				if (send_buf[COMMAND] == REGISTRATION)
				{
					Serial.println(F("Registration error."));
					cancel_registration();
				}
				else
				{
					Serial.println(F("sendtoWait failed."));
				}
			}

			last_registered.clear();
		}
	}
}

void send_to_server(String &data)
{
    WiFiClient client;
    if (WiFi.status() == WL_CONNECTED)
    {
        HTTPClient http;

        http.begin(client, "http://188.35.161.31/core/jsonadd.php");

        http.addHeader("Content-Type", "application/json");
        int httpResponseCode = http.POST(data);

        Serial.print(F("HTTP response code: "));
        Serial.println(httpResponseCode);
        http.end();
    }
    else
    {
        Serial.println(F("WiFi disconnected."));
    }
}

void unpack_telemetry()
{
    String result = "";
    result += '{';
	for (int i = START_PAYLOAD; i < recv_buf[BYTE_COUNT];)
	{
		uint8_t key_id = recv_buf[i++];
		if (key_id == JSON_LEVEL_UP)
		{
			result += '}';
            if (i != recv_buf[BYTE_COUNT] && recv_buf[i] != JSON_LEVEL_UP)
            {
                result += ", ";
            }
			continue;
		}
		result += '\"' + String(registry[key_id]) + '\"' + " : ";
		uint8_t tag = recv_buf[i++];
		switch (tag & GENERAL_TYPE_MASK)
		{
			case SPECIAL:
			{
				switch (tag & SPECIAL_TYPE_MASK)
				{
					case JSON_OBJECT:
					{
                        result += '{';
						continue;
					}
					case IP_ADDR:
					{
						uint8_t ip[4];
                        char ip_str[16];
                        
						ip[0] = recv_buf[i++];
						ip[1] = recv_buf[i++];
						ip[2] = recv_buf[i++];
						ip[3] = recv_buf[i++];
                        sprintf(ip_str, "%hhu.%hhu.%hhu.%hhu", ip[0], ip[1], ip[2], ip[3]);
						
						result += '\"' + String(ip_str) + '\"';
                        if (i != recv_buf[BYTE_COUNT] && recv_buf[i] != JSON_LEVEL_UP)
                        {
                            result += ", ";
                        }
						continue;
					}
					case MAC_ADDR:
					{
						uint8_t mac[6];
                        char mac_str[18];
                        
						mac[0] = recv_buf[i++];
						mac[1] = recv_buf[i++];
						mac[2] = recv_buf[i++];
						mac[3] = recv_buf[i++];
						mac[4] = recv_buf[i++];
						mac[5] = recv_buf[i++];
						sprintf(mac_str, "%02X:%02X:%02X:%02X:%02X:%02X", mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
                      
						result += '\"' + String(mac_str) + '\"';
                        if (i != recv_buf[BYTE_COUNT] && recv_buf[i] != JSON_LEVEL_UP)
                        {
                            result += ", ";
                        }
						continue;
					}
					case DATE_YYYYMMDD:
					{
						uint16_t year = 0;
						year |= recv_buf[i++];
						year |= recv_buf[i++] << 8;
						uint8_t mon = recv_buf[i++];
						uint8_t day = recv_buf[i++];
						char date[11];
						sprintf(date, "%d-%02hhu-%02hhu", year, mon, day);
						
						result += '\"' + String(date) + '\"';
                        if (i != recv_buf[BYTE_COUNT] && recv_buf[i] != JSON_LEVEL_UP)
                        {
                            result += ", ";
                        }
						continue;
					}
					case TIME_HHMMSS:
					{
						uint8_t hour = recv_buf[i++];
						uint8_t min = recv_buf[i++];
						uint8_t sec = recv_buf[i++];
                        char time[11];
                        sprintf(time, "%02hhu:%02hhu:%02hhu", hour, min, sec);

                        result += '\"' + String(time) + '\"';
                        if (i != recv_buf[BYTE_COUNT] && recv_buf[i] != JSON_LEVEL_UP)
                        {
                            result += ", ";
                        }
						continue;
					}
				}
			}
			case STRING:
			{
                strcpy(unpack_buf_cstr, (char*)&recv_buf[i]);
                i += strlen((char*)&recv_buf[i]) + 1;
                
				result += '\"' + String(unpack_buf_cstr) + '\"';
                if (i != recv_buf[BYTE_COUNT] && recv_buf[i] != JSON_LEVEL_UP)
                {
                    result += ", ";
                }
				continue;
			}
			case INTEGER:
			{
				switch (tag & NUMBER_SIZE_MASK)
				{
					case _8BIT:
					{
						int8_t num = (int8_t)recv_buf[i++];

						result += '\"' + String(num) + '\"';
                        if (i != recv_buf[BYTE_COUNT] && recv_buf[i] != JSON_LEVEL_UP)
                        {
                            result += ", ";
                        }
						continue;
					}
					case _16BIT:
					{
						int16_t num = 0;
						num |= recv_buf[i++];
						num |= recv_buf[i++] << 8;

                        result += '\"' + String(num) + '\"';
                        if (i != recv_buf[BYTE_COUNT] && recv_buf[i] != JSON_LEVEL_UP)
                        {
                            result += ", ";
                        }
						continue;
					}
					case _32BIT:
					{
						int32_t num = 0;
						num |= recv_buf[i++];
						num |= recv_buf[i++] << 8;
						num |= recv_buf[i++] << 16;
						num |= recv_buf[i++] << 24;

                        result += '\"' + String(num) + '\"';
                        if (i != recv_buf[BYTE_COUNT] && recv_buf[i] != JSON_LEVEL_UP)
                        {
                            result += ", ";
                        }
						continue;
					}
				}
			}
			case CUSTOM_FLOAT:
			{
				switch (tag & NUMBER_SIZE_MASK)
				{
					case _8BIT:
					{
						int8_t num = (int8_t)recv_buf[i++];
						uint8_t dec_places = tag & DECIMAL_POINT_MASK;
						int8_t integer = num / quick_pow10(dec_places);
						uint8_t fraction = abs(num) % quick_pow10(dec_places);
						char num_str[18];
						sprintf(num_str, "%d.%0*d", integer, dec_places, fraction);
						if (integer == 0 && num < 0)
						{
							result += '-';
						}

                        result += String(num_str);
                        if (i != recv_buf[BYTE_COUNT] && recv_buf[i] != JSON_LEVEL_UP)
                        {
                            result += ", ";
                        }
						continue;
					}
					case _16BIT:
					{
						int16_t num = 0;
						num |= recv_buf[i++];
						num |= recv_buf[i++] << 8;
						uint8_t dec_places = tag & DECIMAL_POINT_MASK;
						int16_t integer = num / quick_pow10(dec_places);
						uint16_t fraction = abs(num) % quick_pow10(dec_places);
						char num_str[18];
						sprintf(num_str, "%d.%0*d", integer, dec_places, fraction);
						if (integer == 0 && num < 0)
						{
							result += '-';
						}

                        result += String(num_str);
                        if (i != recv_buf[BYTE_COUNT] && recv_buf[i] != JSON_LEVEL_UP)
                        {
                            result += ", ";
                        }
						continue;
					}
					case _32BIT:
					{
						int32_t num = 0;
						num |= recv_buf[i++];
						num |= recv_buf[i++] << 8;
						num |= recv_buf[i++] << 16;
						num |= recv_buf[i++] << 24;
						uint8_t dec_places = tag & DECIMAL_POINT_MASK;
						int32_t integer = num / quick_pow10(dec_places);
						uint32_t fraction = abs(num) % quick_pow10(dec_places);
						char num_str[18];
						sprintf(num_str, "%ld.%0*ld", integer, dec_places, fraction);
						if (integer == 0 && num < 0)
						{
							result += '-';
						}

                        result += String(num_str);
                        if (i != recv_buf[BYTE_COUNT] && recv_buf[i] != JSON_LEVEL_UP)
                        {
                            result += ", ";
                        }
						continue;
					}
				}
			}
		}
	}

    result += '}';
    Serial.println(result);

    send_to_server(result);
}

int64_t quick_pow10(int n)
{
	static int64_t pow10[16] =
	{
		1, 10, 100, 1000,
		10000, 100000, 1000000, 10000000,
		1000000000, 10000000000, 100000000000, 1000000000000,
		10000000000000, 100000000000000, 1000000000000000, 10000000000000000
	};

	return pow10[n];
}
